/*
 * Copyright (c) 2014-2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Sep 23 2014
 */
#include "acheader.h"

#include <stdlib.h>
#include <string.h>

#include <avr/cpufunc.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#include "tk3-mikrokopter.h"
#include "mkbl.h"

/* default settings, stored in the MCU EEPROM. The format is a list
 * of nul-terminated strings, interpreted as a key/value/descr */
static const char tk3_dparams[] EEMEM __attribute__((used)) = {
  TK3SET_MAGIC "\0"
  TK3SET_MOTORID "\0"	"\0\0\0\0"		"1-4 or 0 for jumpers" "\0"
  TK3SET_ROTATION "\0"	"\0\0\0\x1"		"+1/-1 trigonometric" "\0"
  TK3SET_POLES "\0"	"\0\0\0\xe"		"motor # poles" "\0"
  TK3SET_MCURRENT "\0"	"\0\0\x23\x28"		"current limit, mA" "\0"
  TK3SET_POST "\0"	"\0\0\0\x1"		"power on self tests" "\0"
  TK3SET_UART "\0"	"\0\x7\xa1\x20"		"uart baud rate" "\0"
  TK3SET_CLKADJ "\0"	"\0\0\0\0"		"internal oscilator offset" "\0"
};

struct tk3_settings settings;

static void	tk3_load_settings(struct tk3_settings *s);

/* intialization tones */
static const uint16_t tones[] = {
  523, 587, 659, 698, 783, 880, 987
};


/* --- main ---------------------------------------------------------------- */

int
main()
{
  tk3_time date;
  struct tk3_iorecv *msg;

  /* setup clock & interrupts */
  if (tk3_clock_init(settings.clk_adj, settings.clk_adj_addr))
    dim_red_led_abort();

  /* red led on */
  DDRD |= (1 << DDD7);
  PORTD |= (1 << PD7);

  /* load settings */
  tk3_load_settings(&settings);

  if (!settings.motor_id) {
    int16_t addr = -1;

    /* I2C address via board jumbers */
    DDRB &= ~((1 << DDB0) | (1 << DDB6) | (1 << DDB7));
    PORTB |= ((1 << PB0) | (1 << PB6) | (1 << PB7));
    while(addr != PINB) addr = PINB; /* synchronize pin output */

    settings.motor_id = 1 +
      ((!(addr & (1 << PINB0)) << 2) |
       (!(addr & (1 << PINB6)) << 1) |
       (!(addr & (1 << PINB7)) << 0));
    PORTB = 0;

    if (settings.motor_id_addr) /* update eeprom */
      tk3_settings_update(settings.motor_id_addr, settings.motor_id);
  }

  /* initialize hardware */
  if (tk3_uart_init(0, settings.uart))
    dim_red_led_abort();
  tk3_twi_init(TK3TWI_BASEADDR + settings.motor_id);

  if (tk3_sensor_init())
    dim_red_led_abort();

  if (tk3_motor_init(settings.rotation))
    dim_red_led_abort();
  if (tk3_motor_test())
    dim_red_led_abort();

  /* red led off */
  DDRD |= (1 << DDD7);
  PORTD &= ~(1 << PD7);

  /* wait 2s and do hardware testing */
  if (settings.post) {
    date = 2000000 + 300000 * settings.motor_id;
    tk3_clock_delay(date);

    if (tk3_motor_tone(tones[settings.motor_id-1]) ||
        tk3_motor_tone(tones[settings.motor_id]))
      dim_red_led_abort();
  }

  /* main loop */
  while(1) {
    date = tk3_clock_gettime();

    if (!tk3_motor.starting && !tk3_motor.spinning)
      tk3_sensor_update(-1);

    /* data logging */
    tk3_log_motor(date);

    /* process received messages */
    while ((msg = tk3_recv())) {
      uint8_t cmd = msg->data[0];
      switch(cmd) {
        case 't': /* clock tick - µs */
          if (msg->len == 5) {
            tk3_time d =
              ((uint32_t)msg->data[1] << 24) |
              ((uint32_t)msg->data[2] << 16) |
              ((uint32_t)msg->data[3] << 8) |
              msg->data[4];

            tk3_clock_reftick(msg->ts, d);
          }
          break;

        case 'p': /* PWM */
          if (msg->len == 3) {
            uint16_t p =
              ((uint16_t)msg->data[1] << 8) |
              ((uint16_t)msg->data[2]);

            tk3_motor_pwm(p);
          }
          break;

        case 'v': /* revolution period - µs */
          if (msg->len == 3) {
            uint16_t t =
              ((uint16_t)msg->data[1] << 8) |
              ((uint16_t)msg->data[2]);

            tk3_motor_period(t);
          }
          break;

        case 'w': /* revolution period array - µs */
          if (msg->len >= 1 + 2 * settings.motor_id) {
            uint16_t t =
              ((uint16_t)msg->data[2 * settings.motor_id - 1] << 8) |
              ((uint16_t)msg->data[2 * settings.motor_id]);

            tk3_motor_period(t);
          }
          break;

        case 'g': /* startup */
          if (!tk3_motor.spinning) {
            if (tk3_motor_start()) dim_red_led_abort();
          }
          break;

        case 'x': /* stop */
          tk3_motor_stop();
          break;

        case 's': /* log motor velocity */
          tk3_log_velocity(msg->state.channel);
          break;

        case 'a': /* log motor current */
          tk3_log_current(msg->state.channel);
          break;

        case 'm': /* log motor data */
          if (msg->len == 5) {
            tk3_time p =
              ((uint32_t)msg->data[1] << 24) |
              ((uint32_t)msg->data[2] << 16) |
              ((uint32_t)msg->data[3] << 8) |
              msg->data[4];

            tk3_logdef_motor(msg->state.channel, p, date);
          }
          break;

        case 'k': /* log controller */
          tk3_log_control(msg->state.channel);
          break;

        case 'd': /* log sensors */
          tk3_log_sensors(msg->state.channel);
          break;

        case '?': /* id */
          if (msg->len == 1) {
            tk3_log(msg->state.channel,
                    PSTR("?%1mkbl" PACKAGE_VERSION), settings.motor_id);
          }
          break;
      }
    }

    /* flush pending data */
    tk3_send();

    /* dim red led blink for motor emergency */
    if (tk3_motor.emerg) {
      if (date & (1UL<<19)) /* ~524ms */
        DDRD |= (1 << DDD7);
      else
        DDRD &= ~(1 << DDD7);
    }
  }

  return 0;
}


/* --- load_settings ------------------------------------------------------- */

static void
tk3_load_settings(struct tk3_settings *s)
{
  int32_t value;
  char key[16];

  if (tk3_settings_init()) dim_red_led_abort();

  while(!tk3_settings_next(key, sizeof(key), &value)) {
    if (!strcmp(key, TK3SET_MOTORID)) {
      s->motor_id = value;

    } else if (!strcmp_P(key, PSTR(TK3SET_ROTATION))) {
      if (value > 0)
        s->rotation = MKBL_ANTICLOCKWISE;
      else if (value < 0)
        s->rotation = MKBL_CLOCKWISE;
      else
        dim_red_led_abort();

    } else if (!strcmp_P(key, PSTR(TK3SET_MCURRENT))) {
      if (value > 0 && value < 15000)
        s->mcurrent = value;
      else
        dim_red_led_abort();

    } else if (!strcmp_P(key, PSTR(TK3SET_POLES))) {
      if (value > 0)
        s->motor_period_mul = value / 2 * 6;
      else
        dim_red_led_abort();

    } else if (!strcmp_P(key, PSTR(TK3SET_POST))) {
      s->post = !!value;

    } else if (!strcmp_P(key, PSTR(TK3SET_UART))) {
      s->uart = value;

    } else if (!strcmp_P(key, PSTR(TK3SET_CLKADJ))) {
      s->clk_adj = value;

    } else
      dim_red_led_abort();
  }

  s->motor_id_addr = tk3_settings_addr(TK3SET_MOTORID);
  s->clk_adj_addr = tk3_settings_addr(TK3SET_CLKADJ);
}


/* --- tk3_comm_overflow --------------------------------------------------- */

void
tk3_comm_overflow(void)
{
  /* red led on */
  DDRD &= ~(1 << DDD7);
}

void
tk3_comm_noverflow(void)
{
  /* red led off */
  DDRD |= (1 << DDD7);
}


/* --- *_abort ------------------------------------------------------------- */

void
red_led_abort(void)
{
  DDRD |= (1 << DDD7);
  PORTD |= (1 << PD7);
  abort();
}

void
flash_red_led_abort(void)
{
  tk3_time t;

  DDRD |= (1 << DDD7);

  t = tk3_clock_gettime();
  while(1) {
    PORTD ^= (1 << PD7);

    t += 500000;
    tk3_clock_delay(t);
  }
}

void
dim_red_led_abort(void)
{
  DDRD &= ~(1 << DDD7);
  PORTD &= ~(1 << PD7);
  abort();
}
